/**
* Copyright 2005-2007 ECMWF
*
* Licensed under the GNU Lesser General Public License which
* incorporates the terms and conditions of version 3 of the GNU
* General Public License.
* See LICENSE and gpl-3.0.txt for details.
*/

#include "grib_api_internal.h"
#include <ctype.h>

/*
   This is used by make_class.pl

   START_CLASS_DEF
   CLASS      = accessor
   SUPER      = grib_accessor_class_codetable
   IMPLEMENTS = init;unpack_string;pack_string
   IMPLEMENTS = unpack_long;pack_long
   IMPLEMENTS = dump
   MEMBERS    =  const char* area
   END_CLASS_DEF

 */

/* START_CLASS_IMP */

/*

Don't edit anything between START_CLASS_IMP and END_CLASS_IMP
Instead edit values between START_CLASS_DEF and END_CLASS_DEF
or edit "accessor.class" and rerun ./make_class.pl

*/

static int pack_long(grib_accessor*, const long* val,size_t *len);
static int pack_string(grib_accessor*, const char*, size_t *len);
static int unpack_long(grib_accessor*, long* val,size_t *len);
static int unpack_string (grib_accessor*, char*, size_t *len);
static void dump(grib_accessor*, grib_dumper*);
static void init(grib_accessor*,const long, grib_arguments* );
static void init_class(grib_accessor_class*);

typedef struct grib_accessor_domain {
    grib_accessor          att;
/* Members defined in gen */
/* Members defined in long */
/* Members defined in unsigned */
	long nbytes;
	grib_arguments* arg;
/* Members defined in codetable */
	const char* tablename;
/* Members defined in domain */
	const char* area;
} grib_accessor_domain;

extern grib_accessor_class* grib_accessor_class_codetable;

static grib_accessor_class _grib_accessor_class_domain = {
    &grib_accessor_class_codetable,                      /* super                     */
    "domain",                      /* name                      */
    sizeof(grib_accessor_domain),  /* size                      */
    0,                           /* inited */
    &init_class,                 /* init_class */
    &init,                       /* init                      */
    0,                  /* post_init                      */
    0,                    /* free mem                       */
    &dump,                       /* describes himself         */
    0,                /* get length of section     */
    0,                /* get number of values      */
    0,                 /* get number of bytes      */
    0,                /* get offset to bytes           */
    0,            /* get native type               */
    0,                /* get sub_section                */
    0,               /* grib_pack procedures long      */
    0,               /* grib_pack procedures long      */
    &pack_long,                  /* grib_pack procedures long      */
    &unpack_long,                /* grib_unpack procedures long    */
    0,                /* grib_pack procedures double    */
    0,              /* grib_unpack procedures double  */
    &pack_string,                /* grib_pack procedures string    */
    &unpack_string,              /* grib_unpack procedures string  */
    0,                 /* grib_pack procedures bytes     */
    0,               /* grib_unpack procedures bytes   */
    0,            /* pack_expression */
    0,              /* notify_change   */
    0,                /* update_size   */
    0,            /* preferred_size   */
    0,                    /* resize   */
    0,      /* nearest_smaller_value */
    0,                       /* next accessor    */
    0,                    /* compare vs. another accessor   */
    0,             /* unpack only ith value          */
};


grib_accessor_class* grib_accessor_class_domain = &_grib_accessor_class_domain;


static void init_class(grib_accessor_class* c)
{
	c->next_offset	=	(*(c->super))->next_offset;
	c->value_count	=	(*(c->super))->value_count;
	c->byte_count	=	(*(c->super))->byte_count;
	c->byte_offset	=	(*(c->super))->byte_offset;
	c->get_native_type	=	(*(c->super))->get_native_type;
	c->sub_section	=	(*(c->super))->sub_section;
	c->pack_missing	=	(*(c->super))->pack_missing;
	c->is_missing	=	(*(c->super))->is_missing;
	c->pack_double	=	(*(c->super))->pack_double;
	c->unpack_double	=	(*(c->super))->unpack_double;
	c->pack_bytes	=	(*(c->super))->pack_bytes;
	c->unpack_bytes	=	(*(c->super))->unpack_bytes;
	c->pack_expression	=	(*(c->super))->pack_expression;
	c->notify_change	=	(*(c->super))->notify_change;
	c->update_size	=	(*(c->super))->update_size;
	c->preferred_size	=	(*(c->super))->preferred_size;
	c->resize	=	(*(c->super))->resize;
	c->nearest_smaller_value	=	(*(c->super))->nearest_smaller_value;
	c->next	=	(*(c->super))->next;
	c->compare	=	(*(c->super))->compare;
	c->unpack_double_element	=	(*(c->super))->unpack_double_element;
}
/* END_CLASS_IMP */


static grib_codetable* load_table(grib_accessor_domain* self)
{
  size_t size = 0;
  grib_handle*    h = ((grib_accessor*)self)->parent->h;
  grib_context*   c = h->context;
  grib_codetable* t = c->codetable;
  FILE *f = NULL;
  char line[1024];
  char filename[1024]={0,};
  int i=0;

  grib_recompose_name(h, NULL,self->tablename, line,1);
  i=0;
  while (i<GRIB_MAX_DEFINITION_DIRS && *filename==0 )
      grib_context_full_path(c,line,filename,i++);


  while(t)
  {
    if(strcmp(filename,t->filename) == 0)
      return t;
    t = t->next;
  }

  f=fopen(filename, "r");


  if (!f)
  {
    grib_context_log(c,(GRIB_LOG_WARNING)|(GRIB_LOG_PERROR),"code_table_entry: cannot open code table %s",filename);
    return NULL;
  }

  size = 512;

  t = (grib_codetable*)grib_context_malloc_clear_persistent(c,sizeof(grib_codetable) + (size-1)*sizeof(code_table_entry));
  t->filename  = grib_context_strdup_persistent(c,filename);
  t->next      = c->codetable;
  t->size      = size;
  c->codetable = t;

  while(fgets(line,sizeof(line)-1,f))
  {
    const char* p = line;
    int code    = 0;
    char abbreviation[1024] = {0,};
    char* q = abbreviation;
    line[strlen(line)-1] = 0;

    while(*p && isspace(*p))
      p++;

    if(*p == '#')
      continue;

    while(*p && isspace(*p))
      p++;

    while(*p)
    {
      if(isspace(*p))
        break;
      code *= 10;
      code += *p - '0';
      p++;
    }

    if(code <0 || code >= size)
    {
      grib_context_log(c,(GRIB_LOG_WARNING)|(GRIB_LOG_PERROR)," %s: %d (table size=%d)",filename,code,size);
      continue;
    }

    while(*p && isspace(*p))
      p++;

    while(*p)
    {
      if(isspace(*p))
        break;
      *q++ = *p++;
    }
    *q = 0;
    while(*p && isspace(*p))
      p++;

    t->entries[code].abbreviation = grib_context_strdup_persistent(c,abbreviation);
    t->entries[code].title        = grib_context_strdup_persistent(c,p);

  }

  fclose(f);

  return t;

}

static void init(grib_accessor* a, const long len, grib_arguments* params)
{
  grib_accessor_domain* self = (grib_accessor_domain*)a;
  self->area = grib_arguments_get_name(a->parent->h,params,1);
  a->length = 0;
}

static void dump(grib_accessor* a, grib_dumper* dumper)
{
  grib_accessor_domain* self = (grib_accessor_domain*)a;
  grib_codetable*          table = load_table(self);
  char comment[1024];
  int ret = GRIB_SUCCESS;
  size_t llen = 1;
  long value;

  ret = grib_unpack_long(a, &value,&llen);
  if(ret == GRIB_SUCCESS)
  {

    if(table && value >= 0 && value < table->size && table->entries[value].title)
    {
      sprintf(comment,"%s",table->entries[value].title);
    }
    else
    {
      strcpy(comment,"Unknown domain, assumed global");
    }
  }
  else
  {
    strcpy(comment,"Undefined");
  }
  grib_dump_string(dumper,a,comment);

}

static int unpack_string (grib_accessor* a, char* val, size_t *len)
{

  int ret = GRIB_SUCCESS;
  grib_accessor_class* super = *(a->cclass->super);

  ret = super->unpack_string(a,val,len);

  if (ret ==  GRIB_SUCCESS)
    return GRIB_SUCCESS;


  if(*len < 2)
  {
    *len = 2;
    return GRIB_BUFFER_TOO_SMALL;
  }

  strcpy(val,"g");
  *len = 2;

  return GRIB_SUCCESS;
}

static int pack_string(grib_accessor* a, const char* buffer, size_t *len)
{
  return GRIB_NOT_IMPLEMENTED;
}

#define PRECISION 0.1

static int equal_area(float a, double b)
{
  if(abs(a-b) > PRECISION)
    return 0;
  return 1;
}

static int unpack_long(grib_accessor* a, long* val,size_t *len)
{
  grib_accessor_domain* self = (grib_accessor_domain*)a;
  grib_codetable*          table = load_table(self);
  size_t i = 0;

  int err = GRIB_SUCCESS;

  double area[4];

  float north = 0;
  float south = 0;
  float east  = 0;
  float west  = 0;

  size_t size = 4;



  if(!table)
  {
    *val = 0;
    *len = 1;
    return GRIB_DECODING_ERROR;
  }


  if( (err = grib_get_double_array_internal(a->parent->h,self->area,area,&size)) != GRIB_SUCCESS)
  {
    *val = 0;
    *len = 1;
    return err;
  }

  for(i = 0;i < table->size; i++)
  {
    if(table->entries[i].title)
    {

      sscanf(table->entries[i].title,"%g %g %g %g",&north,&west,&south,&east);
      if(equal_area(north,area[0])
          && equal_area(west,area[1])
          && equal_area(south,area[2])
          && equal_area(east,area[3]))
      {

        *len = 1;
        *val = i;
        return GRIB_SUCCESS;
      }
    }
  }
  *val = 0;
  *len = 1;
  return err;
  return GRIB_DECODING_ERROR;

}
static int pack_long(grib_accessor* a, const long* val,size_t *len)
{
  return GRIB_NOT_IMPLEMENTED;
}
