/**
* Copyright 2005-2007 ECMWF
*
* Licensed under the GNU Lesser General Public License which
* incorporates the terms and conditions of version 3 of the GNU
* General Public License.
* See LICENSE and gpl-3.0.txt for details.
*/

/*********************************************
 *   Enrico Fucile
 *******************************************/

#include "grib_api_internal.h"
/*
   This is used by make_class.pl

   START_CLASS_DEF
   CLASS      = accessor
   SUPER      = grib_accessor_class_ascii
   IMPLEMENTS = pack_string;unpack_string;value_count
   IMPLEMENTS = pack_long
   IMPLEMENTS = init
   MEMBERS = const char* start
   MEMBERS = const char* end
   MEMBERS = const char* timeRangeIndicator
   MEMBERS = const char* unit
   END_CLASS_DEF

 */

/* START_CLASS_IMP */

/*

Don't edit anything between START_CLASS_IMP and END_CLASS_IMP
Instead edit values between START_CLASS_DEF and END_CLASS_DEF
or edit "accessor.class" and rerun ./make_class.pl

*/

static int pack_long(grib_accessor*, const long* val,size_t *len);
static int pack_string(grib_accessor*, const char*, size_t *len);
static int unpack_string (grib_accessor*, char*, size_t *len);
static long value_count(grib_accessor*);
static void init(grib_accessor*,const long, grib_arguments* );
static void init_class(grib_accessor_class*);

typedef struct grib_accessor_g1marsstep {
    grib_accessor          att;
/* Members defined in gen */
/* Members defined in ascii */
/* Members defined in g1marsstep */
	const char* start;
	const char* end;
	const char* timeRangeIndicator;
	const char* unit;
} grib_accessor_g1marsstep;

extern grib_accessor_class* grib_accessor_class_ascii;

static grib_accessor_class _grib_accessor_class_g1marsstep = {
    &grib_accessor_class_ascii,                      /* super                     */
    "g1marsstep",                      /* name                      */
    sizeof(grib_accessor_g1marsstep),  /* size                      */
    0,                           /* inited */
    &init_class,                 /* init_class */
    &init,                       /* init                      */
    0,                  /* post_init                      */
    0,                    /* free mem                       */
    0,                       /* describes himself         */
    0,                /* get length of section     */
    &value_count,                /* get number of values      */
    0,                 /* get number of bytes      */
    0,                /* get offset to bytes           */
    0,            /* get native type               */
    0,                /* get sub_section                */
    0,               /* grib_pack procedures long      */
    0,               /* grib_pack procedures long      */
    &pack_long,                  /* grib_pack procedures long      */
    0,                /* grib_unpack procedures long    */
    0,                /* grib_pack procedures double    */
    0,              /* grib_unpack procedures double  */
    &pack_string,                /* grib_pack procedures string    */
    &unpack_string,              /* grib_unpack procedures string  */
    0,                 /* grib_pack procedures bytes     */
    0,               /* grib_unpack procedures bytes   */
    0,            /* pack_expression */
    0,              /* notify_change   */
    0,                /* update_size   */
    0,            /* preferred_size   */
    0,                    /* resize   */
    0,      /* nearest_smaller_value */
    0,                       /* next accessor    */
    0,                    /* compare vs. another accessor   */
    0,             /* unpack only ith value          */
};


grib_accessor_class* grib_accessor_class_g1marsstep = &_grib_accessor_class_g1marsstep;


static void init_class(grib_accessor_class* c)
{
	c->dump	=	(*(c->super))->dump;
	c->next_offset	=	(*(c->super))->next_offset;
	c->byte_count	=	(*(c->super))->byte_count;
	c->byte_offset	=	(*(c->super))->byte_offset;
	c->get_native_type	=	(*(c->super))->get_native_type;
	c->sub_section	=	(*(c->super))->sub_section;
	c->pack_missing	=	(*(c->super))->pack_missing;
	c->is_missing	=	(*(c->super))->is_missing;
	c->unpack_long	=	(*(c->super))->unpack_long;
	c->pack_double	=	(*(c->super))->pack_double;
	c->unpack_double	=	(*(c->super))->unpack_double;
	c->pack_bytes	=	(*(c->super))->pack_bytes;
	c->unpack_bytes	=	(*(c->super))->unpack_bytes;
	c->pack_expression	=	(*(c->super))->pack_expression;
	c->notify_change	=	(*(c->super))->notify_change;
	c->update_size	=	(*(c->super))->update_size;
	c->preferred_size	=	(*(c->super))->preferred_size;
	c->resize	=	(*(c->super))->resize;
	c->nearest_smaller_value	=	(*(c->super))->nearest_smaller_value;
	c->next	=	(*(c->super))->next;
	c->compare	=	(*(c->super))->compare;
	c->unpack_double_element	=	(*(c->super))->unpack_double_element;
}
/* END_CLASS_IMP */

static void init(grib_accessor* a,const long l, grib_arguments* c)
{
  grib_accessor_g1marsstep* self = (grib_accessor_g1marsstep*)a;
  grib_handle* h=a->parent->h;
  int n=0;
  a->length=0;

  self->start   = grib_arguments_get_name(h,c,n++);
  self->end   = grib_arguments_get_name(h,c,n++);
  self->timeRangeIndicator   = grib_arguments_get_name(h,c,n++);
  self->unit   = grib_arguments_get_name(h,c,n++);
}

static int pack_string(grib_accessor* a, const char* val, size_t *len){
  grib_accessor_g1marsstep* self = (grib_accessor_g1marsstep*)a;
  grib_handle* h=a->parent->h;
  long timeRangeIndicator=0;
  long start=0,end=-1,unit=0,ounit=0;
  int unit_value=1;
  int ret=0;
  char *p=NULL,*q=NULL;

  if((ret = grib_get_long_internal(h,self->timeRangeIndicator,&timeRangeIndicator)))
      return ret;

  if (self->unit)
    if((ret = grib_get_long_internal(h,self->unit,&unit)))
      return ret;

  ounit=unit;
  start=strtol(val, &p,10);

  switch (timeRangeIndicator) {
    case 2:
    case 3:
    case 4:
    case 5:
      if (*p==0) return GRIB_WRONG_STEP;
  }

  end=-1;
  if ( *p!=0 ) end=strtol(++p, &q,10);

  unit=1;
  if (self->unit && end!=-1 && (start>255 || end>255) ) {
    grib_g1_step_apply_units(&start,&end,&unit,&unit_value);
  }

  if (self->unit && ounit != unit)
    if((ret = grib_set_long_internal(h,self->unit,unit)))
      return ret;

  start*=unit_value;
  if((ret = grib_set_long_internal(h,self->start,start)))
      return ret;

  if (end!=-1) {
    end*=unit_value;
    if((ret = grib_set_long_internal(h,self->end,end)))
      return ret;
  }

  return 0;

}

static int    unpack_string(grib_accessor* a, char* val, size_t *len)
{
  grib_accessor_g1marsstep* self = (grib_accessor_g1marsstep*)a;
  grib_handle* h=a->parent->h;
  char buf[100];
  int ret=0;
  size_t size=0;

  long start=0,end=0;

  if((ret = grib_get_long_internal(h,self->start,&start)))
      return ret;

  if((ret = grib_get_long_internal(h,self->end,&end)))
      return ret;

  if (start==end || start==0)
    sprintf(buf,"%ld",end);
  else
    sprintf(buf,"%ld-%ld",start,end);

  size=strlen(buf)+1;

  if (*len<size) return GRIB_ARRAY_TOO_SMALL;

  *len=size;

  memcpy(val,buf,size);

  return GRIB_SUCCESS;
}


static long value_count(grib_accessor* a)
{
  char result[1024]  ;
  size_t s = sizeof(result);

  unpack_string(a,result,&s);

    return s;
}

static int pack_long(grib_accessor* a, const long* val, size_t *len)
{
  char buff[100];
  size_t bufflen=100;
  sprintf(buff,"%ld",*val);
  return pack_string( a,buff,&bufflen);
}

