/**
* Copyright 2005-2007 ECMWF
*
* Licensed under the GNU Lesser General Public License which
* incorporates the terms and conditions of version 3 of the GNU
* General Public License.
* See LICENSE and gpl-3.0.txt for details.
*/

/***************************************
 *  Enrico Fucile
 **************************************/

#include "grib_api_internal.h"
/*
   This is used by make_class.pl

   START_CLASS_DEF
   CLASS      = accessor
   SUPER      = grib_accessor_class_unsigned
   IMPLEMENTS = unpack_long;pack_long;value_count
   IMPLEMENTS = init;dump
   MEMBERS    = const char* p1
   MEMBERS    = const char* p2
   MEMBERS    = const char *indicator
   MEMBERS    = const char *unit
   END_CLASS_DEF

 */

/* START_CLASS_IMP */

/*

Don't edit anything between START_CLASS_IMP and END_CLASS_IMP
Instead edit values between START_CLASS_DEF and END_CLASS_DEF
or edit "accessor.class" and rerun ./make_class.pl

*/

static int pack_long(grib_accessor*, const long* val,size_t *len);
static int unpack_long(grib_accessor*, long* val,size_t *len);
static long value_count(grib_accessor*);
static void dump(grib_accessor*, grib_dumper*);
static void init(grib_accessor*,const long, grib_arguments* );
static void init_class(grib_accessor_class*);

typedef struct grib_accessor_g1step {
    grib_accessor          att;
/* Members defined in gen */
/* Members defined in long */
/* Members defined in unsigned */
	long nbytes;
	grib_arguments* arg;
/* Members defined in g1step */
	const char* p1;
	const char* p2;
	const char *indicator;
	const char *unit;
} grib_accessor_g1step;

extern grib_accessor_class* grib_accessor_class_unsigned;

static grib_accessor_class _grib_accessor_class_g1step = {
    &grib_accessor_class_unsigned,                      /* super                     */
    "g1step",                      /* name                      */
    sizeof(grib_accessor_g1step),  /* size                      */
    0,                           /* inited */
    &init_class,                 /* init_class */
    &init,                       /* init                      */
    0,                  /* post_init                      */
    0,                    /* free mem                       */
    &dump,                       /* describes himself         */
    0,                /* get length of section     */
    &value_count,                /* get number of values      */
    0,                 /* get number of bytes      */
    0,                /* get offset to bytes           */
    0,            /* get native type               */
    0,                /* get sub_section                */
    0,               /* grib_pack procedures long      */
    0,               /* grib_pack procedures long      */
    &pack_long,                  /* grib_pack procedures long      */
    &unpack_long,                /* grib_unpack procedures long    */
    0,                /* grib_pack procedures double    */
    0,              /* grib_unpack procedures double  */
    0,                /* grib_pack procedures string    */
    0,              /* grib_unpack procedures string  */
    0,                 /* grib_pack procedures bytes     */
    0,               /* grib_unpack procedures bytes   */
    0,            /* pack_expression */
    0,              /* notify_change   */
    0,                /* update_size   */
    0,            /* preferred_size   */
    0,                    /* resize   */
    0,      /* nearest_smaller_value */
    0,                       /* next accessor    */
    0,                    /* compare vs. another accessor   */
    0,             /* unpack only ith value          */
};


grib_accessor_class* grib_accessor_class_g1step = &_grib_accessor_class_g1step;


static void init_class(grib_accessor_class* c)
{
	c->next_offset	=	(*(c->super))->next_offset;
	c->byte_count	=	(*(c->super))->byte_count;
	c->byte_offset	=	(*(c->super))->byte_offset;
	c->get_native_type	=	(*(c->super))->get_native_type;
	c->sub_section	=	(*(c->super))->sub_section;
	c->pack_missing	=	(*(c->super))->pack_missing;
	c->is_missing	=	(*(c->super))->is_missing;
	c->pack_double	=	(*(c->super))->pack_double;
	c->unpack_double	=	(*(c->super))->unpack_double;
	c->pack_string	=	(*(c->super))->pack_string;
	c->unpack_string	=	(*(c->super))->unpack_string;
	c->pack_bytes	=	(*(c->super))->pack_bytes;
	c->unpack_bytes	=	(*(c->super))->unpack_bytes;
	c->pack_expression	=	(*(c->super))->pack_expression;
	c->notify_change	=	(*(c->super))->notify_change;
	c->update_size	=	(*(c->super))->update_size;
	c->preferred_size	=	(*(c->super))->preferred_size;
	c->resize	=	(*(c->super))->resize;
	c->nearest_smaller_value	=	(*(c->super))->nearest_smaller_value;
	c->next	=	(*(c->super))->next;
	c->compare	=	(*(c->super))->compare;
	c->unpack_double_element	=	(*(c->super))->unpack_double_element;
}
/* END_CLASS_IMP */

static void init(grib_accessor* a,const long l, grib_arguments* c)
{
  grib_accessor_g1step* self = (grib_accessor_g1step*)a;
  int n = 0;
  self->p1        = grib_arguments_get_name(a->parent->h,c,n++);
  self->p2        = grib_arguments_get_name(a->parent->h,c,n++);
  self->indicator = grib_arguments_get_name(a->parent->h,c,n++);
  self->unit      = grib_arguments_get_name(a->parent->h,c,n++);
}

static void dump(grib_accessor* a, grib_dumper* dumper)
{
  grib_dump_string(dumper,a,NULL);

}
static int units[] =  {
  0,      /* (0) minutes */
  1,      /* (1) hour    */
  24,     /* (2) day     <------- modified for UK seasonal forecast
         use of time-range indicator = 2 */
  0, 0, 0, 0, 0, 0, 0,

  3,     /* (10) 3 hours */
  6,     /* (11) 6 hours */
  12,    /* (12) 12 hours */
};

static int units_index[] = {
   1,10,11,12,2
  };

int grib_g1_step_get_steps(grib_accessor* a,long* start,long* end, long *indicator)
{
  grib_accessor_g1step* self = (grib_accessor_g1step*)a;
  int err = 0;
  long p1 = 0,p2 = 0,unit = 0;

  err = grib_get_long_internal(a->parent->h,self->unit,&unit);
  if(err)           return err;

  err = grib_get_long_internal(a->parent->h,self->p1,&p1);
  if(err)               return err;

  err = grib_get_long_internal(a->parent->h,self->p2,&p2);
  if(err)               return err;

  err = grib_get_long_internal(a->parent->h,self->indicator,indicator);
  if(err)  return err;

  *start = p1;
  *end   = p2;

  switch(*indicator)
  {
    case 10:
      *start = *end = (p1<<8) | (p2<<0);
      break;

    case 113:
    case 114:
    case 115:
    case 116:
    case 117:
    case 118:
    case 119:
    case 123:
    case 124:
      *start = *end = p1;
      break;

  }
  if(*start > *end)
    *end = *start;

  *start = (*start) * units[unit];
  *end   = (*end)   * units[unit];
  return 0;
}

int grib_g1_step_apply_units(long *val,long *sval2,long *unit,int *unit_value) {
  int i=0,j=0,found=0;
  long val2=0;

  if (*val<255 && *unit==1)
    return 0;

  found=0;
  *unit_value=0;
  for (i=0;i<5 ;i++) {
   if (units_index[i] == *unit) {found=1;break;}
  }

  if (!found) return GRIB_WRONG_STEP;

  val2=*sval2*units[units_index[i]];

  for (j=i;j<5;j++) {
    /*printf("units=%d val=%ld val2=%ld\n",units[units_index[j]],*val,val2);*/
    if (*val%units[units_index[j]]==0 && val2%units[units_index[j]]==0
       && ((*val/units[units_index[j]])<255) && ((val2/units[units_index[j]])< 255)) {
      *val=*val/units[units_index[j]];
      *sval2=val2/units[units_index[j]];
      *unit=units_index[j];
      *unit_value=units[units_index[j]];
      return 0;
    }
  }

  if (i==0) return GRIB_WRONG_STEP;

  for (j=0;j<i;j++) {
    if (*val%units[units_index[j]]==0 && val2%units[units_index[j]]==0
       && ((*val/units[units_index[j]])<255) && ((val2/units[units_index[j]])< 255)) {
      *val=*val/units[units_index[j]];
      *sval2=val2/units[units_index[j]];
      *unit=units_index[j];
      return 0;
    }
  }

  return GRIB_INTERNAL_ERROR;
}

static int unpack_long(grib_accessor* a, long* val, size_t *len)
{
  long start = 0, end = 0;
  long indicator = 0;

  int err = grib_g1_step_get_steps(a,&start,&end, &indicator);

  if(err) return err;

  *val = (long)end;

  return GRIB_SUCCESS;
}

static int pack_long(grib_accessor* a, const long* val, size_t *len)
{
  grib_accessor_g1step* self = (grib_accessor_g1step*)a;

  long lval=*val;
  int unit_value=0;
  int err = 0;
  long p1 = 0,p2 = 0,unit = 0,indicator=0;
  long op1 = 0,op2 = 0,ounit = 0,off=0;
  long zero=0;
  grib_accessor* p1_accessor=0;


  err = grib_get_long_internal(a->parent->h,self->unit,&unit);
  if(err) return err;

  err = grib_get_long_internal(a->parent->h,self->p1,&p1);
  if(err) return err;

  err = grib_get_long_internal(a->parent->h,self->p2,&p2);
  if(err) return err;

  err = grib_get_long_internal(a->parent->h,self->indicator,&indicator);
  if(err) return err;

  op1=p1;
  op2=p2;
  ounit=unit;

  switch (indicator) {
    case 0:
      err=grib_g1_step_apply_units(&lval,&zero,&unit,&unit_value);
      if (err) return err;
      err = grib_set_long_internal(a->parent->h,self->p1,lval);
      if (err) return err;
      err = grib_set_long_internal(a->parent->h,self->p2,0);
      if (err) return err;
      break;

    case 1:
      indicator=0;
      err = grib_set_long_internal(a->parent->h,self->indicator,indicator);
      if(err) return err;
      err = grib_set_long_internal(a->parent->h,self->p1,lval);
      return 0;

    case 2:
    case 3:
    case 4:
    case 5:
      err=grib_g1_step_apply_units(&lval,&p1,&unit,&unit_value);
      if (err) return err;

      if (op1 != p1)
        err = grib_set_long_internal(a->parent->h,self->p1,p1);
      if (err) return err;

      err = grib_set_long_internal(a->parent->h,self->p2,lval);
      if (err) return err;
      break;

    case 10:
      unit=1;
      p1_accessor=grib_find_accessor( a->parent->h,self->p1);
      off = p1_accessor->offset*8;
      err = grib_encode_unsigned_long(a->parent->h->buffer->data, lval,&off,16);
      if (err == GRIB_SUCCESS) len[0] = 1;
      break;

    case 113:
    case 114:
    case 115:
    case 116:
    case 117:
    case 118:
    case 119:
    case 123:
    case 124:
      err=grib_g1_step_apply_units(&lval,&p2,&unit,&unit_value);
      if (err) return err;

      if (op2 != p2)
        err = grib_set_long_internal(a->parent->h,self->p2,p2);
      if (err) return err;

      err = grib_set_long_internal(a->parent->h,self->p1,lval);
      if (err) return err;

      break;

    default:
        return GRIB_NOT_IMPLEMENTED;
  }

  if (unit != ounit)
    err = grib_set_long_internal(a->parent->h,self->unit,unit);
  if (err) return err;

  return err;
}

static long value_count(grib_accessor* a){  return 1;}

int grib_g1_time_units(int *tunits,long code) {
  int err=0,i=0;
  int found=0;

  for (i=0;i<5 ;i++) {
   if (units_index[i] == code) {
      found=1;
      *tunits=units[units_index[i]];
      break;
    }
  }

  if (!found) return GRIB_WRONG_STEP;



  return err;

}

