/**
* Copyright 2005-2007 ECMWF
*
* Licensed under the GNU Lesser General Public License which
* incorporates the terms and conditions of version 3 of the GNU
* General Public License.
* See LICENSE and gpl-3.0.txt for details.
*/

#include "grib_api_internal.h"
/*
   This is used by make_class.pl

   START_CLASS_DEF
   CLASS      = accessor
   SUPER      = grib_accessor_class_long
   IMPLEMENTS = unpack_long;pack_long
   IMPLEMENTS = init;dump
MEMBERS = const char* year
MEMBERS = const char* month
MEMBERS = const char* day
MEMBERS = const char* hour
MEMBERS = const char* minute
MEMBERS = const char* second
MEMBERS = const char* forecast_time
MEMBERS = const char* unit
MEMBERS = const char* year_of_end_of_interval
MEMBERS = const char* month_of_end_of_interval
MEMBERS = const char* day_of_end_of_interval
MEMBERS = const char* hour_of_end_of_interval
MEMBERS = const char* minute_of_end_of_interval
MEMBERS = const char* second_of_end_of_interval
MEMBERS = const char* n
MEMBERS = const char* type_of_increment
MEMBERS = const char* process_unit
MEMBERS = const char* process_time_range
MEMBERS = const char* increment_unit
MEMBERS = const char* increment_time_range
   END_CLASS_DEF

 */

/* START_CLASS_IMP */

/*

Don't edit anything between START_CLASS_IMP and END_CLASS_IMP
Instead edit values between START_CLASS_DEF and END_CLASS_DEF
or edit "accessor.class" and rerun ./make_class.pl

*/

static int pack_long(grib_accessor*, const long* val,size_t *len);
static int unpack_long(grib_accessor*, long* val,size_t *len);
static void dump(grib_accessor*, grib_dumper*);
static void init(grib_accessor*,const long, grib_arguments* );
static void init_class(grib_accessor_class*);

typedef struct grib_accessor_g2end_step {
    grib_accessor          att;
/* Members defined in gen */
/* Members defined in long */
/* Members defined in g2end_step */
	const char* year;
	const char* month;
	const char* day;
	const char* hour;
	const char* minute;
	const char* second;
	const char* forecast_time;
	const char* unit;
	const char* year_of_end_of_interval;
	const char* month_of_end_of_interval;
	const char* day_of_end_of_interval;
	const char* hour_of_end_of_interval;
	const char* minute_of_end_of_interval;
	const char* second_of_end_of_interval;
	const char* n;
	const char* type_of_increment;
	const char* process_unit;
	const char* process_time_range;
	const char* increment_unit;
	const char* increment_time_range;
} grib_accessor_g2end_step;

extern grib_accessor_class* grib_accessor_class_long;

static grib_accessor_class _grib_accessor_class_g2end_step = {
    &grib_accessor_class_long,                      /* super                     */
    "g2end_step",                      /* name                      */
    sizeof(grib_accessor_g2end_step),  /* size                      */
    0,                           /* inited */
    &init_class,                 /* init_class */
    &init,                       /* init                      */
    0,                  /* post_init                      */
    0,                    /* free mem                       */
    &dump,                       /* describes himself         */
    0,                /* get length of section     */
    0,                /* get number of values      */
    0,                 /* get number of bytes      */
    0,                /* get offset to bytes           */
    0,            /* get native type               */
    0,                /* get sub_section                */
    0,               /* grib_pack procedures long      */
    0,               /* grib_pack procedures long      */
    &pack_long,                  /* grib_pack procedures long      */
    &unpack_long,                /* grib_unpack procedures long    */
    0,                /* grib_pack procedures double    */
    0,              /* grib_unpack procedures double  */
    0,                /* grib_pack procedures string    */
    0,              /* grib_unpack procedures string  */
    0,                 /* grib_pack procedures bytes     */
    0,               /* grib_unpack procedures bytes   */
    0,            /* pack_expression */
    0,              /* notify_change   */
    0,                /* update_size   */
    0,            /* preferred_size   */
    0,                    /* resize   */
    0,      /* nearest_smaller_value */
    0,                       /* next accessor    */
    0,                    /* compare vs. another accessor   */
    0,             /* unpack only ith value          */
};


grib_accessor_class* grib_accessor_class_g2end_step = &_grib_accessor_class_g2end_step;


static void init_class(grib_accessor_class* c)
{
	c->next_offset	=	(*(c->super))->next_offset;
	c->value_count	=	(*(c->super))->value_count;
	c->byte_count	=	(*(c->super))->byte_count;
	c->byte_offset	=	(*(c->super))->byte_offset;
	c->get_native_type	=	(*(c->super))->get_native_type;
	c->sub_section	=	(*(c->super))->sub_section;
	c->pack_missing	=	(*(c->super))->pack_missing;
	c->is_missing	=	(*(c->super))->is_missing;
	c->pack_double	=	(*(c->super))->pack_double;
	c->unpack_double	=	(*(c->super))->unpack_double;
	c->pack_string	=	(*(c->super))->pack_string;
	c->unpack_string	=	(*(c->super))->unpack_string;
	c->pack_bytes	=	(*(c->super))->pack_bytes;
	c->unpack_bytes	=	(*(c->super))->unpack_bytes;
	c->pack_expression	=	(*(c->super))->pack_expression;
	c->notify_change	=	(*(c->super))->notify_change;
	c->update_size	=	(*(c->super))->update_size;
	c->preferred_size	=	(*(c->super))->preferred_size;
	c->resize	=	(*(c->super))->resize;
	c->nearest_smaller_value	=	(*(c->super))->nearest_smaller_value;
	c->next	=	(*(c->super))->next;
	c->compare	=	(*(c->super))->compare;
	c->unpack_double_element	=	(*(c->super))->unpack_double_element;
}
/* END_CLASS_IMP */

static void init(grib_accessor* a,const long l, grib_arguments* c)
{
  grib_accessor_g2end_step* self = (grib_accessor_g2end_step*)a;
  int n = 0;

  self->year  = grib_arguments_get_name(a->parent->h,c,n++);
  self->month = grib_arguments_get_name(a->parent->h,c,n++);
  self->day = grib_arguments_get_name(a->parent->h,c,n++);
  self->hour  = grib_arguments_get_name(a->parent->h,c,n++);
  self->minute  = grib_arguments_get_name(a->parent->h,c,n++);
  self->second  = grib_arguments_get_name(a->parent->h,c,n++);
  self->forecast_time = grib_arguments_get_name(a->parent->h,c,n++);
  self->unit  = grib_arguments_get_name(a->parent->h,c,n++);
  self->year_of_end_of_interval = grib_arguments_get_name(a->parent->h,c,n++);
  self->month_of_end_of_interval  = grib_arguments_get_name(a->parent->h,c,n++);
  self->day_of_end_of_interval  = grib_arguments_get_name(a->parent->h,c,n++);
  self->hour_of_end_of_interval = grib_arguments_get_name(a->parent->h,c,n++);
  self->minute_of_end_of_interval = grib_arguments_get_name(a->parent->h,c,n++);
  self->second_of_end_of_interval = grib_arguments_get_name(a->parent->h,c,n++);
  self->n = grib_arguments_get_name(a->parent->h,c,n++);
  self->type_of_increment = grib_arguments_get_name(a->parent->h,c,n++);
  self->process_unit  = grib_arguments_get_name(a->parent->h,c,n++);
  self->process_time_range  = grib_arguments_get_name(a->parent->h,c,n++);
  self->increment_unit  = grib_arguments_get_name(a->parent->h,c,n++);
  self->increment_time_range  = grib_arguments_get_name(a->parent->h,c,n++);

}

static void dump(grib_accessor* a, grib_dumper* dumper)
{
  grib_dump_double(dumper,a,NULL);

}
static int units[] =  {
  0,      /* (0) minutes */
  1,      /* (1) hour    */
  24,     /* (2) day */

  0, 0, 0, 0, 0, 0, 0,

  3,     /* (10) 3 hours */
  6,     /* (11) 6 hours */
  12,    /* (12) 12 hours */
};

static int unpack_long(grib_accessor* a, long* val, size_t *len)
{
  grib_accessor_g2end_step* self = (grib_accessor_g2end_step*)a;
  int err = 0;

  long minute;
  long second;

  long forecast_time;
  long unit;

  long minute_of_end_of_interval;
  long second_of_end_of_interval;
  long n;
  long type_of_increment;

  long process_unit;
  long process_time_range;
  long increment_unit;
  long increment_time_range;


  if((err = grib_get_long_internal(a->parent->h,self->minute,&minute))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->second,&second))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->forecast_time,&forecast_time))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->unit,&unit))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->minute_of_end_of_interval,&minute_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->second_of_end_of_interval,&second_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->n,&n))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->type_of_increment,&type_of_increment))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->process_unit,&process_unit))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->process_time_range,&process_time_range))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->increment_unit,&increment_unit))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->increment_time_range,&increment_time_range))) return err;


  if(type_of_increment != 2)
  {
    printf("AAAAA type_of_increment %ld != 2\n", type_of_increment);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(!units[unit])
  {
    printf("AAAAA bad unit %ld\n",unit);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(!units[process_unit])
  {
    printf("AAAAA bad process_unit %ld\n",process_unit);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(n != 1)
  {
    printf("AAAAA n != 1\n");
    return GRIB_NOT_IMPLEMENTED;
  }

/*
This is the sampling of the data it doesn't affect the end of the interval

  if(increment_time_range != 0)
  {
    printf("AAAAA increment_time_range %ld != 0\n",increment_time_range);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(increment_unit != 255 && increment_time_range != 0)
  {
    printf("AAAAA increment_unit(%ld) != 255\n",increment_unit);
    return GRIB_NOT_IMPLEMENTED;
  }

*/
/*
  if(minute_of_end_of_interval != 0) {
    printf("AAAAA minute_of_end_of_interval %ld != 0\n",minute_of_end_of_interval);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(second_of_end_of_interval != 0) {
    printf("AAAAA second_of_end_of_interval %ld != 0\n",second_of_end_of_interval);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(minute != 0) {
    printf("AAAAA minute %ld != 0\n",minute);
    return GRIB_NOT_IMPLEMENTED;
  }

  if(second != 0) {
    printf("AAAAA second %ld != 0\n",second);
    return GRIB_NOT_IMPLEMENTED;
  }
*/

  /*
  printf("------- forecast_time=%ld units=%d\n",forecast_time,units[unit]);
  printf("------- process_time_range=%ld units=%d\n",process_time_range,units[process_unit]);
  */

  forecast_time      *= units[unit];
  process_time_range *= units[process_unit];

#if 0
  hours = (grib_date_to_julian(year_of_end_of_interval*10000 + month_of_end_of_interval*100 + day_of_end_of_interval) -
           grib_date_to_julian(year*10000 + month*100 + day)) * 24;

  hours += (hour_of_end_of_interval - hour);
  hours += process_time_range;
#endif

  *val = forecast_time + process_time_range;


  return GRIB_SUCCESS;
}

static int pack_long(grib_accessor* a, const long* val, size_t *len)
{
  grib_accessor_g2end_step* self = (grib_accessor_g2end_step*)a;

  int err = 0;

  long year;
  long month;
  long day;
  long hour;
  long minute;
  long second;

  long forecast_time;
  long unit;
  long year_of_end_of_interval;
  long month_of_end_of_interval;
  long day_of_end_of_interval;
  long hour_of_end_of_interval;
  long minute_of_end_of_interval = 0;
  long second_of_end_of_interval = 0;
  long n;
  long type_of_increment;
  long process_unit = 1;
  long process_time_range;
  long increment_unit = 255;
  long increment_time_range = 0;

  long end;
  long date;
  int end_of_interval_assigned=0,time_length_assigned=0,increment_time_assigned=0;

  if((err = grib_get_long_internal(a->parent->h,self->year,&year))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->month,&month))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->day,&day))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->hour,&hour))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->minute,&minute))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->second,&second))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->forecast_time,&forecast_time))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->unit,&unit))) return err;

 if((err = grib_get_long_internal(a->parent->h,self->year_of_end_of_interval,&year_of_end_of_interval))) return err;
  if (year_of_end_of_interval !=0 ) end_of_interval_assigned=1;
#if 0
  if((err = grib_get_long_internal(a->parent->h,self->year_of_end_of_interval,&year_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->month_of_end_of_interval,&month_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->day_of_end_of_interval,&day_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->hour_of_end_of_interval,&hour_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->minute_of_end_of_interval,&minute_of_end_of_interval))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->second_of_end_of_interval,&second_of_end_of_interval))) return err;
#endif

  if((err = grib_get_long_internal(a->parent->h,self->n,&n))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->type_of_increment,&type_of_increment))) return err;

  if((err = grib_get_long_internal(a->parent->h,self->process_time_range,&process_time_range))) return err;
  if (process_time_range!=0) time_length_assigned=1;

  if((err = grib_get_long_internal(a->parent->h,self->increment_unit,&increment_unit))) return err;
  if (increment_unit!=0) increment_time_assigned=1;
#if 0
  if((err = grib_get_long_internal(a->parent->h,self->process_unit,&process_unit))) return err;
  if((err = grib_get_long_internal(a->parent->h,self->increment_time_range,&increment_time_range))) return err;
#endif

  if(n != 1)
  {
    printf("ZZZZZZ n != 1\n");
    return GRIB_NOT_IMPLEMENTED;
  }

  if(type_of_increment != 2)
  {
    printf("ZZZZZZ type_of_increment != 2\n");
    return GRIB_NOT_IMPLEMENTED;
  }

  if(minute != 0) {
    printf("ZZZZZZ minute != 0\n");
    return GRIB_NOT_IMPLEMENTED;
  }
  if(second != 0) {
    printf("ZZZZZZ second != 0\n");
    return GRIB_NOT_IMPLEMENTED;
  }

  if(!units[unit])
  {
    printf("ZZZZZZ bad unit %ld\n",unit);
    return GRIB_NOT_IMPLEMENTED;
  }


  if(!units[process_unit])
  {
    printf("ZZZZZZ bad process_unit %ld\n",process_unit);
    return GRIB_NOT_IMPLEMENTED;
  }

  forecast_time      *= units[unit];
  process_time_range  = *val;
  Assert(process_time_range>=0);

  process_time_range -= forecast_time;

  Assert(process_time_range>=0);



  /* printf("QQQQQ forecast_time %ld process_time_range %ld\n", */
    /* forecast_time,process_time_range); */


  end  = grib_date_to_julian(year*10000 + month*100 + day) * 24 + hour;
  end += process_time_range + forecast_time;
  date = grib_julian_to_date(end/24);

  hour_of_end_of_interval = end % 24;

  year_of_end_of_interval  =  date/10000;
  month_of_end_of_interval = (date%10000)/100;
  day_of_end_of_interval   =  date%100;

  if (!end_of_interval_assigned) {
    if((err = grib_set_long_internal(a->parent->h,self->year_of_end_of_interval,  year_of_end_of_interval))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->month_of_end_of_interval, month_of_end_of_interval))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->day_of_end_of_interval,   day_of_end_of_interval))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->hour_of_end_of_interval,  hour_of_end_of_interval))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->minute_of_end_of_interval,minute_of_end_of_interval))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->second_of_end_of_interval,second_of_end_of_interval))) return err;
  }
  if (!time_length_assigned) {
    if((err = grib_set_long_internal(a->parent->h,self->process_unit,process_unit))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->process_time_range,process_time_range))) return err;
  }
  if (!increment_time_assigned) {
    if((err = grib_set_long_internal(a->parent->h,self->increment_unit,increment_unit))) return err;
    if((err = grib_set_long_internal(a->parent->h,self->increment_time_range,increment_time_range))) return err;
  }

  return GRIB_SUCCESS;
}

