/**
* Copyright 2005-2007 ECMWF
*
* Licensed under the GNU Lesser General Public License which
* incorporates the terms and conditions of version 3 of the GNU
* General Public License.
* See LICENSE and gpl-3.0.txt for details.
*/

/***************************************************************************
 *   Enrico Fucile  - 19.06.2007                                           *
 *                                                                         *
 ***************************************************************************/
#include "grib_api_internal.h"

#if OMP_PACKING
#include "omp.h"
#endif

#define mask1(i)    (1u << i)
#define test(n,i)    !!((n) & mask1(i))

long MASK = -1;                             /* Mask of sword bits */

# define VALUE(p,q,b) \
 (((b)==max_nbits ? MASK : ~(MASK<<(b))) & ((p)>>(max_nbits-((q)+(b)))))

# define MASKVALUE(q,b) \
 ((b)==max_nbits ? MASK : (~(MASK<<(b))<<(max_nbits-((q)+(b)))))

/*
static unsigned long *bmasks=0;

static void init_bmasks() {
  if (!bmasks) {
    int i,wordBits=sizeof(long)*8;
    unsigned long cmask=-1;
    bmasks=(unsigned long*)malloc(sizeof(long)*wordBits);
    for (i=1;i < wordBits; i++) {
      bmasks[i-1] = ~(cmask << i);
    }
    bmasks[wordBits-1] = cmask << wordBits;
  }
}
static int max_nbytes = sizeof(unsigned long);
*/

static unsigned long dmasks[] = { 0xFF, 0xFE, 0xFC, 0xF8, 0xF0, 0xE0, 0xC0, 0x80, 0x00, };

static int max_nbits = sizeof(unsigned long)*8;

static void init_ibm_table();

typedef struct ibm_table_t ibm_table_t;

struct ibm_table_t {
  int inited;
  double e[128];
  double v[128];
  double vmin;
  double vmax;
};

static ibm_table_t ibm_table={ 0,{0,},{0,} };

static void init_ibm_table() {
  if (!ibm_table.inited) {
    unsigned long i;
    unsigned long mmin=0x100000;
    unsigned long mmax=0xffffff;
    double e=1;
    for (i=1; i<=57;i++) { e*=16; ibm_table.e[i+70]=e;ibm_table.v[i+70]=e*mmin;}
    ibm_table.e[70]=1;
    ibm_table.v[70]=mmin;
    e=1;
    for (i=1; i<=70;i++) { e/=16; ibm_table.e[-i+70]=e;ibm_table.v[-i+70]=e*mmin;}
    ibm_table.vmin=ibm_table.v[0];
    ibm_table.vmax=ibm_table.e[127]*mmax;
    ibm_table.inited=1;
    /*for (i=0;i<128;i++) printf("++++ ibm_table.v[%d]=%g\n",i,ibm_table.v[i]);*/
  }
}


static void binary_search(double xx[], const unsigned long n, double x, unsigned long *j)
{
  /*These routine works only on ascending ordered arrays*/
  unsigned long ju,jm,jl;
  jl=0;
  ju=n;
  while (ju-jl > 1) {
    jm=(ju+jl) >> 1;
    /* printf("jl=%lu jm=%lu ju=%lu\n",jl,jm,ju); */
    /* printf("xx[jl]=%.10e xx[jm]=%.10e xx[ju]=%.10e\n",xx[jl],xx[jm],xx[ju]); */
    if (x >= xx[jm]) jl=jm;
    else ju=jm;
  }
  *j=jl;
}

unsigned long grib_ibm_to_long(double x) {
  unsigned long s = 0;
  unsigned long mmax = (1<<24)-1;
  unsigned long mmin = 0x100000;
  unsigned long m = mmax;
  unsigned long e=0;
  double rmmax=mmax+0.5;

  if (!ibm_table.inited) init_ibm_table();

  /* printf("\ngrib_ibm_to_long: x=%.20e\n",x); */
  if (x < 0)  {  s  = 1; x = -x; }

  /* Underflow */
  if (x < ibm_table.vmin) {
    /*printf("grib_ibm_to_long: (x < ibm_table.vmin) x=%.20e vmin=%.20e v=0x%lX\n",x,ibm_table.vmin,(s<<31));*/
    return (s << 31);
  }

  /* Overflow */
  Assert(x <= ibm_table.vmax);
/*
  if (x > ibm_table.vmax) {
    printf("grib_ibm_to_long: x=%.20e > xmax=%.20e\n",x,ibm_table.vmax);
  return 0;
  }
*/

  binary_search(ibm_table.v, 127, x, &e);

  /* printf("grib_ibm_to_long: e=%ld\n",e); */

  x/=ibm_table.e[e];

  /* printf("grib_ibm_to_long: x=%.20e\n",x); */

  while(x < mmin ) { x *= 16; e--;
    /* printf("grib_ibm_to_long (e--): x=%.20e e=%ld \n",x,e); */
  }

  while(x > rmmax ) { x /= 16; e++;
    /* printf("grib_ibm_to_long (e++): x=%.20e e=%ld \n",x,e); */
  }

  m=x+0.5;
  /* printf("grib_ibm_to_long: m=0x%lX (%lu) x=%.10e \n",m,m,x ); */
  if ( m > mmax ) { e++; m=0x100000;
     /* printf("grib_ibm_to_long: ( m > mmax ) m=0x%lX (%lu) x=%.10e \n",m,m,x ); */
  }

  /* printf("grib_ibm_to_long: s=%lu c=%lu (0x%lX) m=%lu (0x%lX)\n",s,e,e,m,m ); */

  return (s << 31) | ( e << 24 ) | m;
}

double grib_long_to_ibm(unsigned long x){
  unsigned long s = x  & 0x80000000;
  unsigned long c = (x & 0x7f000000) >> 24;
  unsigned long m = (x & 0x00ffffff);

  double val = m;

  if (!ibm_table.inited) init_ibm_table();

  if(x == 0) return 0;

  val*=ibm_table.e[c];

  if(s) val = -val;

  return val;
}

double grib_ibm_table_e(unsigned long e) {
  if (!ibm_table.inited) init_ibm_table();
  return ibm_table.e[e];
}

double grib_ibm_table_v(unsigned long e) {
  if (!ibm_table.inited) init_ibm_table();
  return ibm_table.v[e];
}

unsigned long grib_ibm_nearest_smaller_to_long(double x)
{
   unsigned long l;
   unsigned long e;
   unsigned long m ;
   unsigned long s;
   unsigned long mmin = 0x100000;
   double y,eps;

   if(x == 0) return 0;

   if (!ibm_table.inited) init_ibm_table();

   l=grib_ibm_to_long(x);
   y=grib_long_to_ibm(l);

   if ( x < y ) {
     if ( x < 0 && -x < ibm_table.vmin ) {
       l=0x80100000;
     } else {
       e = (l & 0x7f000000) >> 24;
       m = (l & 0x00ffffff);
       s  = l  & 0x80000000;

       if ( m == mmin ) {
         /* printf("grib_ibm_nearest_smaller_to_long: m == mmin (0x%lX) e=%lu\n",m,e);  */
   e = s ? e : e-1;
   if (e<0) e=0;
   if (e>127) e=127;
   /* printf("grib_ibm_nearest_smaller_to_long: e=%lu \n",e);  */
       }

       eps=ibm_table.e[e];

       /* printf("grib_ibm_nearest_smaller_to_long: x<y\n"); */
       l=grib_ibm_to_long(y-eps);
       /* printf("grib_ibm_nearest_smaller_to_long: grib_ibm_to_long(y-eps)=0x%lX y=%.10e eps=%.10e x=%.10e\n",l,y,eps,x); */
     }
   }

   if (x<grib_long_to_ibm(l)) {
     printf("grib_ibm_nearest_smaller_to_long: x=%.20e grib_long_to_ibm(0x%lX)=%.20e\n",x,l,grib_long_to_ibm(l));
     Assert(x >= grib_long_to_ibm(l));
   }

   return l;
}

int grib_nearest_smaller_ibm_float(double a,double* ret) {
  unsigned long l=0;

  if (!ibm_table.inited) init_ibm_table();

  if (a>ibm_table.vmax) return GRIB_INTERNAL_ERROR;

  l=grib_ibm_nearest_smaller_to_long(a);
  *ret=grib_long_to_ibm(l);
  return GRIB_SUCCESS;
}

double grib_nearest_smaller_ieee_float(double a)
{
  double e =  grib_long_to_ieee(grib_ieee_to_long(a));
  double b = a;

  /* printf("----> a=%g e=%g e-a=%g\n",a,e,e-a); */

  if( e > b )
  {
    unsigned long ub = grib_ieee_to_long(b);
    unsigned long ue;
    while(e>b)
    {
      /* printf("a=%g e=%g e-a=%g\n",a,e,e-a); */
      a -= (e-a);
      e = grib_long_to_ieee(grib_ieee_to_long(a));
    }
    ue = grib_ieee_to_long(e);
    Assert((ue-ub) == 1);
  }

  Assert(b >= e);
  return e;
}

double grib_long_to_ieee(unsigned long x)
{
  unsigned long s = x  & 0x80000000;
  unsigned long c = (x & 0x7f800000) >> 23;
  unsigned long m;

  double val;
  long e;

  if(x == 0)
    return 0;

  Assert(c != 255);

  if(c == 0)
  {
    m = x & 0x007fffff;
    e = -126 - 23;
  }
  else
  {
    m = (x & 0x007fffff) | (1<<23);
    e = c - 127 - 23;
  }


  val = m;

  while(e < 0)
  {
    val /= 2.0;
    e++;
  }

  while(e > 0)
  {
    val *= 2.0;
    e--;
  }

  if(s)
    val = -val;


  return val;
}

unsigned long grib_ieee_to_long(double x)
{
  /* double y = x; */
  unsigned long s = 0;
  unsigned long m;
  long p = 0;
  unsigned long e = 0;

  if(x == 0)
    return 0;

  if(x < 0) { s  = 1; x = -x; }
  while(x < 2) { x *= 2; p--; }

  while(x >= 2) { x /= 2; p++; }

  if(p > 127 ) {
    /* Overflow */
    e = 255;
    m = 0;
  }
  else if(p < -126) {
    /* int i; */
    e = 0;
    /* printf("p=%ld x=%g %ld\n",p,x,p+126+23); */
    m = x * grib_power(p+126+23,2);
  }
  else {
    e = p + 127;
    m = x * ( 1 << 23);
    m &= 0x007fffff;
  }

  m =  (s << 31) | ( e << 23 ) | m;

  return m;
}

unsigned long grib_decode_unsigned_byte_long(const unsigned char* p, long o, int l)
{
  long accum = 0;
  int i = 0;
  unsigned char b = p[o++];

  Assert(l <= max_nbits);

  accum  <<= 8;
  accum |= b ;

  for ( i=1; i<l; i++ )
  {
    b = p[o++];
    accum <<= 8;
    accum |= b ;
  }
  return accum;
}

long grib_decode_signed_long(const unsigned char* p, long o, int l)
{
  long accum = 0;
  int i = 0;
  unsigned char b = p[o++];
  int sign = grib_get_bit(&b, 0);

  Assert(l <= max_nbits);

  b &= 0x7f;
  accum  <<= 8;
  accum |= b ;

  for ( i=1; i<l; i++ )
  {
    b = p[o++];
    accum <<= 8;
    accum |= b ;
  }
  if (sign == 0)
    return accum;
  else
    return -accum;
}

int grib_encode_signed_long(unsigned char* p, long val , long o, int l)
{
  unsigned short accum = 0;
  int i = 0;
  int off = o;
  int sign = (val<0);

  Assert(l <= max_nbits);

  if (sign) val *= -1;

  for (  i=0; i<l; i++ )
  {
    accum    =  val >> (l*8-(8*(i+1)));
    p[o++] =  accum ;
  }

  if (sign) p[off] |= 128;

  return GRIB_SUCCESS;
}

static void grib_set_bit_on( unsigned char* p, long* bitp){

  p +=  *bitp/8;
  *p |=  (1u << (7-((*bitp)%8)));
  (*bitp)++;
}

static void grib_set_bit_off( unsigned char* p, long* bitp){

  p +=  *bitp/8;
  *p &= ~(1u << (7-((*bitp)%8)));
  (*bitp)++;
}

int grib_get_bit(const unsigned char* p, long bitp){
  p += (bitp >> 3);
  return (*p&(1<<(7-(bitp%8))));
}

void grib_set_bit( unsigned char* p, long bitp, int val){

  if(val == 0) grib_set_bit_off(p,&bitp);
  else grib_set_bit_on(p,&bitp);
}

long grib_decode_signed_longb(const unsigned char* p, long *bitp, long nbits)
{
  int sign = grib_get_bit(p, *bitp);
  long val = 0;

  Assert(nbits <= max_nbits);

  *bitp += 1;

  val = grib_decode_unsigned_long( p, bitp,  nbits-1);

  if (sign)val = -val;


  return  val;
}

#if FAST_BIG_ENDIAN

unsigned long grib_decode_unsigned_long(const unsigned char* p, long *bitp, long nbits)
{
  long countOfLeftmostBits=0,leftmostBits=0;
  long startBit= *bitp ;
  long remainingBits = nbits;
  long *pp=(long*)p;
  unsigned long val=0;

  if (startBit >= max_nbits) {
    pp += startBit/max_nbits;
    startBit %= max_nbits;
  }

  countOfLeftmostBits = startBit + remainingBits;
  if (countOfLeftmostBits > max_nbits) {
    countOfLeftmostBits = max_nbits - startBit;
    remainingBits -= countOfLeftmostBits;
    leftmostBits=(VALUE(*pp,startBit,countOfLeftmostBits)) << remainingBits;
    startBit = 0;
    pp++;
  } else
    leftmostBits = 0;

  val=leftmostBits+(VALUE(*pp,startBit,remainingBits));

  *bitp += nbits;

  return val;
}

#if 1
int grib_decode_double_array(const unsigned char* p, long *bitp, long nbits,double reference_value,double s,double d,size_t size,double* val) {
  long i=0;
  long countOfLeftmostBits=0,leftmostBits=0;
  long startBit,startByte;
  long remainingBits = nbits;
  long *pp=(long*)p;
  int inited=0;
  unsigned long uval=0;
  double fact=s*d;
  double bias=reference_value*d;

  startBit=*bitp;
  remainingBits = nbits;
  if (startBit >= max_nbits) {
    pp+=startBit/max_nbits;
    startBit %= max_nbits;
  }

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {

    for (i=0;i<size;i++) {
      if (startBit == max_nbits) {
        startBit = 0;
        pp++;
      }

      val[i]=VALUE(*pp,startBit,remainingBits);
      val[i]= val[i] * fact + bias ;
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  } else {
    for (i=0;i<size;i++) {
      countOfLeftmostBits = startBit + remainingBits;
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        remainingBits -= countOfLeftmostBits;
        leftmostBits=(VALUE(*(pp++),startBit,countOfLeftmostBits)) << remainingBits;
        startBit = 0;
      } else
        leftmostBits = 0;

      val[i]=leftmostBits+(VALUE(*pp,startBit,remainingBits));
      val[i]= val[i] * fact + bias ;
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  }

  *bitp+=size*nbits;

  return GRIB_SUCCESS;
}

int grib_decode_double_array_complex(const unsigned char* p, long *bitp, long nbits,double reference_value,double s,double* d,size_t size,double* val) {
  long i=0;
  long countOfLeftmostBits=0,leftmostBits=0;
  long startBit;
  long remainingBits = nbits;
  long *pp=(long*)p;
  int inited=0;
  unsigned long uval=0;

  startBit=*bitp;
  remainingBits = nbits;
  if (startBit >= max_nbits) {
     pp+=startBit/max_nbits;
     startBit %= max_nbits;
  }

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {
    for (i=0;i<size;i++) {
      if (startBit == max_nbits) {
        startBit = 0;
        pp++;
      }

      val[i]=VALUE(*pp,startBit,remainingBits);
      val[i]= ((( (val[i]) * s)+reference_value)*d[i/2]);
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  } else {

    for (i=0;i<size;i++) {
      countOfLeftmostBits = startBit + remainingBits;
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        remainingBits -= countOfLeftmostBits;
        leftmostBits=(VALUE(*pp,startBit,countOfLeftmostBits)) << remainingBits;
        startBit = 0;
        pp++;
      } else
        leftmostBits = 0;

      val[i]=leftmostBits+(VALUE(*pp,startBit,remainingBits));
      val[i]= ((( (val[i]) * s)+reference_value)*d[i/2]);
      startBit+=remainingBits;
      remainingBits=nbits;
    }
  }

  *bitp+=size*nbits;

  return GRIB_SUCCESS;
}

#else
int grib_decode_double_array(const unsigned char* p, long *bitp, long nbits,double reference_value,double s,double d,size_t size,double* val) {
  long i=0;
  long countOfLeftmostBits=0,leftmostBits=0;
  long startBit,startByte;
  long remainingBits = nbits;
  long *pp=(long*)p;
  int inited=0;
  unsigned long uval=0;
  double fact=s*d;
  double bias=reference_value*d;

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {

#pragma omp parallel for schedule(static) firstprivate(inited,pp) private(startBit,countOfLeftmostBits,remainingBits,leftmostBits)
    for (i=0;i<size;i++) {

      if (!inited) {
        startBit=*bitp+i*nbits;
        remainingBits = nbits;
        if (startBit >= max_nbits) {
          pp+=startBit/max_nbits;
          startBit %= max_nbits;
        }
        inited=1;
      }

      if (startBit == max_nbits) {
        startBit = 0;
        pp++;
      }

      val[i]=VALUE(*pp,startBit,remainingBits);
      val[i]= val[i] * fact + bias ;
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  } else {

#pragma omp parallel for schedule(static) firstprivate(inited,pp) private(startBit,countOfLeftmostBits,remainingBits,leftmostBits)
    for (i=0;i<size;i++) {

      if (!inited) {
        startBit=*bitp+i*nbits;
        remainingBits = nbits;
        if (startBit >= max_nbits) {
          pp+=startBit/max_nbits;
          startBit %= max_nbits;
        }
        inited=1;
      }

      countOfLeftmostBits = startBit + remainingBits;
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        remainingBits -= countOfLeftmostBits;
        leftmostBits=(VALUE(*(pp++),startBit,countOfLeftmostBits)) << remainingBits;
        startBit = 0;
      } else
        leftmostBits = 0;

      val[i]=leftmostBits+(VALUE(*pp,startBit,remainingBits));
      val[i]= val[i] * fact + bias ;
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  }

  *bitp+=size*nbits;

  return GRIB_SUCCESS;
}

int grib_decode_double_array_complex(const unsigned char* p, long *bitp, long nbits,double reference_value,double s,double* d,size_t size,double* val) {
  long i=0;
  long countOfLeftmostBits=0,leftmostBits=0;
  long startBit;
  long remainingBits = nbits;
  long *pp=(long*)p;
  int inited=0;
  unsigned long uval=0;

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {

#pragma omp parallel for schedule(static) firstprivate(inited,pp) private(startBit,countOfLeftmostBits,remainingBits,leftmostBits)
    for (i=0;i<size;i++) {

      if (!inited) {
        startBit=*bitp+i*nbits;
        remainingBits = nbits;
        if (startBit >= max_nbits) {
          pp+=startBit/max_nbits;
          startBit %= max_nbits;
        }
        inited=1;
      }

      if (startBit == max_nbits) {
        startBit = 0;
        pp++;
      }

      val[i]=VALUE(*pp,startBit,remainingBits);
      val[i]= ((( (val[i]) * s)+reference_value)*d[i/2]);
      startBit+=remainingBits;
      remainingBits=nbits;
    }

  } else {

#pragma omp parallel for schedule(static) firstprivate(inited,pp) private(startBit,countOfLeftmostBits,remainingBits,leftmostBits)
    for (i=0;i<size;i++) {

      if (!inited) {
        startBit=*bitp+i*nbits;
        remainingBits = nbits;
        if (startBit >= max_nbits) {
          pp+=startBit/max_nbits;
          startBit %= max_nbits;
        }
        inited=1;
      }

      countOfLeftmostBits = startBit + remainingBits;
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        remainingBits -= countOfLeftmostBits;
        leftmostBits=(VALUE(*pp,startBit,countOfLeftmostBits)) << remainingBits;
        startBit = 0;
        pp++;
      } else
        leftmostBits = 0;

      val[i]=leftmostBits+(VALUE(*pp,startBit,remainingBits));
      val[i]= ((( (val[i]) * s)+reference_value)*d[i/2]);
      startBit+=remainingBits;
      remainingBits=nbits;
    }
  }

  *bitp+=size*nbits;

  return GRIB_SUCCESS;
}
#endif


#else

unsigned long grib_decode_unsigned_long(const unsigned char* p, long *bitp, long nbits)
{
  int i;
  long ret = 0;
  long o = *bitp/8;
  int l = nbits/8;

  if (nbits==0) return 0;

  if(nbits > max_nbits)
  {
    int bits = nbits;
    int mod = bits % max_nbits;

    if(mod != 0)
    {
      int e=grib_decode_unsigned_long(p,bitp,mod);
      Assert( e == 0);
      bits -= mod;
    }

    while(bits > max_nbits)
    {
      int e=grib_decode_unsigned_long(p,bitp,max_nbits);
      Assert( e == 0);
      bits -= max_nbits;
    }

    return grib_decode_unsigned_long(p,bitp,bits);
  }

  if((nbits%8 > 0)||(*bitp%8 > 0)) {
    for(i=0; i< nbits;i++){
      ret <<= 1;
      if(grib_get_bit( p, *bitp)) ret += 1;
      *bitp += 1;
    }
    return ret;
  }

  ret  <<= 8;
  ret |= p[o++] ;

  for ( i=1; i<l; i++ )
  {
    ret <<= 8;
    ret |= p[o++] ;
  }
  *bitp += nbits;

  return ret;
}

int grib_decode_double_array(const unsigned char* p, long *bitp, long bitsPerValue,
                             double reference_value,double s,double d,
                             size_t n_vals,double* val) {
  long i=0;
  long lvalue = 0;

  if(bitsPerValue%8)
  {
    int j=0;
    for(i=0;i < n_vals;i++) {
      lvalue=0;
      for(j=0; j< bitsPerValue;j++){
        lvalue <<= 1;
        if(grib_get_bit( p, *bitp)) lvalue += 1;
        *bitp += 1;
      }
      val[i] = (double) (((lvalue*s)+reference_value)*d);
    }
  }  else  {
    int bc;
    int l = bitsPerValue/8;
    size_t o = 0;

    for(i=0;i < n_vals;i++)
    {
      lvalue  = 0;
      lvalue  <<= 8;
      lvalue |= p[o++] ;

      for ( bc=1; bc<l; bc++ )
      {
        lvalue <<= 8;
        lvalue |= p[o++] ;
      }
      val[i] = (double) (((lvalue*s)+reference_value)*d);
    }
  }

  return 0;
}


int grib_decode_double_array_complex(const unsigned char* p, long *bitp, long nbits,double reference_value,double s,double* d,size_t size,double* val) {
  return GRIB_NOT_IMPLEMENTED;
}

#endif


int grib_encode_signed_longb(unsigned char* p,  long val ,long *bitp, long nb)
{
  short  sign = val < 0;

  Assert(nb <= max_nbits);

  if(sign) val = -val;

  if(sign)
    grib_set_bit_on (p, bitp);
  else
    grib_set_bit_off(p, bitp);

  return grib_encode_unsigned_longb(p, val ,bitp, nb-1);
}

#if FAST_BIG_ENDIAN
int grib_encode_unsigned_long(unsigned char* p, unsigned long val ,long *bitp, long nbits)
{
  long* destination = (long*)p;
  long countOfLeftmostBits=0,nextWord=0,startBit=0,remainingBits=0,rightmostBits=0;

  startBit = *bitp;
  remainingBits = nbits;

  if (startBit >= max_nbits) {
    nextWord = startBit / max_nbits;
    startBit %= max_nbits;
  } else
    nextWord = 0;

  countOfLeftmostBits = startBit + remainingBits;
  if (countOfLeftmostBits > max_nbits) {
    countOfLeftmostBits = max_nbits - startBit;
    startBit = max_nbits - remainingBits;
    remainingBits -= countOfLeftmostBits;
    destination[nextWord] =
      ((destination[nextWord] >> countOfLeftmostBits) << countOfLeftmostBits)
      + (VALUE(val,startBit,countOfLeftmostBits));
    startBit = 0;
    nextWord++;
  }

  rightmostBits = VALUE(val,max_nbits-remainingBits,remainingBits);
  destination[nextWord] =
    (destination[nextWord] & ~MASKVALUE(startBit,remainingBits))
    + (rightmostBits << max_nbits-(remainingBits+startBit));

  *bitp+=nbits;
  return GRIB_SUCCESS;
}

int grib_encode_unsigned_longb(unsigned char* p, unsigned long val ,long *bitp, long nbits)
{
  return grib_encode_unsigned_long(p,val ,bitp, nbits);
}

int grib_encode_double_array(size_t n_vals,const double* val,long nbits,double reference_value,double d,double divisor,unsigned char* p,long *bitp)
{
  long* destination = (long*)p;
  double* v=(double*)val;
  long countOfLeftmostBits=0,startBit=0,remainingBits=0,rightmostBits=0;
  unsigned long uval=0;
  size_t i=0;

  startBit=*bitp;
  remainingBits = nbits;

  if (startBit >= max_nbits) {
    destination += startBit / max_nbits;
    startBit %= max_nbits;
  }

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {

    for(i=0;i< n_vals;i++){
      uval  = (unsigned long)(((((*v)*d)-reference_value)*divisor)+0.5);
      if (startBit == max_nbits) {
        startBit = 0;
        destination++;
      }

      rightmostBits = VALUE(uval,max_nbits-remainingBits,remainingBits);
      *destination = ((*destination) & ~MASKVALUE(startBit,remainingBits))
        + (rightmostBits << max_nbits-(remainingBits+startBit));
      startBit+=remainingBits;
      remainingBits=nbits;
      v++;
    }

  } else {

    for(i=0;i< n_vals;i++){
      countOfLeftmostBits = startBit + remainingBits;
      uval  = (unsigned long)(((((*v)*d)-reference_value)*divisor)+0.5);
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        startBit = max_nbits - remainingBits;
        remainingBits -= countOfLeftmostBits;
        *destination = (((*destination) >> countOfLeftmostBits) << countOfLeftmostBits)
          + (VALUE(uval,startBit,countOfLeftmostBits));
        startBit = 0;
        destination++;
      }

      rightmostBits = VALUE(uval,max_nbits-remainingBits,remainingBits);
      *destination = ((*destination) & ~MASKVALUE(startBit,remainingBits))
        + (rightmostBits << max_nbits-(remainingBits+startBit));
      startBit+=remainingBits;
      remainingBits=nbits;
      v++;
    }

  }

  *bitp+=n_vals*nbits;

  return GRIB_SUCCESS;
}

int grib_encode_double_array_complex(size_t n_vals,double* val,long nbits,double reference_value,
        double* scal,double d,double divisor,unsigned char* p,long *bitp) {
  long* destination = (long*)p;
  double* v=val;
  long countOfLeftmostBits=0,startBit=0,remainingBits=0,rightmostBits=0;
  unsigned long uval=0;
  size_t i=0;

  startBit=*bitp;
  remainingBits = nbits;

  if (startBit >= max_nbits) {
    destination += startBit / max_nbits;
    startBit %= max_nbits;
  }

  if ( (max_nbits%nbits == 0) && (*bitp%nbits == 0) ) {

    for(i=0;i< n_vals;i++) {
      uval  = (unsigned long)(((((*v)*d*scal[i/2])-reference_value)*divisor)+0.5);
      if (startBit == max_nbits) {
        startBit = 0;
        destination++;
      }

      rightmostBits = VALUE(uval,max_nbits-remainingBits,remainingBits);
      *destination = ((*destination) & ~MASKVALUE(startBit,remainingBits))
        + (rightmostBits << max_nbits-(remainingBits+startBit));
      startBit+=remainingBits;
      remainingBits=nbits;

      v++;
    }

  } else {

    for(i=0;i< n_vals;i++) {
      countOfLeftmostBits = startBit + remainingBits;
      uval  = (unsigned long)(((((*v)*d*scal[i/2])-reference_value)*divisor)+0.5);
      if (countOfLeftmostBits > max_nbits) {
        countOfLeftmostBits = max_nbits - startBit;
        startBit = max_nbits - remainingBits;
        remainingBits -= countOfLeftmostBits;
        *destination = (((*destination) >> countOfLeftmostBits) << countOfLeftmostBits)
          + (VALUE(uval,startBit,countOfLeftmostBits));
        startBit = 0;
        destination++;
      }

      rightmostBits = VALUE(uval,max_nbits-remainingBits,remainingBits);
      *destination = ((*destination) & ~MASKVALUE(startBit,remainingBits))
        + (rightmostBits << max_nbits-(remainingBits+startBit));
      startBit+=remainingBits;
      remainingBits=nbits;

      v++;
    }

  }

  *bitp+=n_vals*nbits;

  return 0;
}
#else

int grib_encode_unsigned_long(unsigned char* p, unsigned long val ,long *bitp, long nbits)
{
  long            len     = nbits;
  int              s      = *bitp%8;
  int              n      = 8-s;

  if(nbits > max_nbits)
  {
    /* TODO: Do some real code here, to support long longs */
    int bits = nbits;
    int mod = bits % max_nbits;
    long zero = 0;

    /* printf("Warning: encoding %ld bits=%ld %ld\n",val,nbits,*bitp); */

    if(mod != 0)
    {
      int e=grib_encode_unsigned_long(p,zero,bitp,mod);
      /* printf(" -> : encoding %ld bits=%ld %ld\n",zero,(long)mod,*bitp); */
      Assert( e == 0);
      bits -= mod;
    }

    while(bits > max_nbits)
    {
      int e=grib_encode_unsigned_long(p,zero,bitp,max_nbits);
      /* printf(" -> : encoding %ld bits=%ld %ld\n",zero,(long)max_nbits,*bitp); */
      Assert(e == 0);
      bits -= max_nbits;
    }

    /* printf(" -> : encoding %ld bits=%ld %ld\n",val,(long)bits,*bitp); */
    return grib_encode_unsigned_long(p,val,bitp,bits);

  }

  if(s)
    p += (*bitp >> 3); /* skip the bytes */
  else
    p += (*bitp >> 3); /* skip the bytes */

  /* head */
  if(s) {
    len   -= n;
    if (len < 0) *p  = ((val << -len) | ((*p)++ & dmasks[n]));
    else *p  = ((val >> len)| ((*p)++ & dmasks[n]));
  }

  /*  write the middle words */
  while(len >= 8)
  {
    len   -= 8;
    *p++ = (val >> len);
  }

  /*  write the end bits */
  if(len) *p =  (val << (8-len));

  *bitp += nbits;
  return GRIB_SUCCESS;
}

int grib_encode_unsigned_longb(unsigned char* p, unsigned long val ,long *bitp, long nb)
{
  long  i = 0;

  Assert(nb <= max_nbits);

  for(i=nb-1; i >= 0; i--){
    if(test(val,i))
      grib_set_bit_on (p, bitp);
    else
      grib_set_bit_off(p, bitp);
  }
  return GRIB_SUCCESS;
}

int grib_encode_double_array(size_t n_vals,const double* val,long bits_per_value,double reference_value,double d,double divisor,unsigned char* p,long *off)
{
  size_t i=0;
  unsigned long unsigned_val=0;
  unsigned char *encoded=p;
  if(bits_per_value%8){
    for(i=0;i< n_vals;i++){
      unsigned_val = (unsigned long)((((val[i]*d)-reference_value)*divisor)+0.5);
      grib_encode_unsigned_longb(encoded, unsigned_val, off , bits_per_value);
    }
  } else{
    for(i=0;i< n_vals;i++){
    int blen=0;
      blen = bits_per_value;
      unsigned_val = (unsigned long)((((val[i]*d)-reference_value)*divisor)+0.5);
      while(blen >= 8)
      {
        blen   -= 8;
        *encoded = (unsigned_val >> blen);
        encoded++;
        *off+=8;
      }
    }
  }
  return GRIB_SUCCESS;
}

int grib_encode_double_array_complex(size_t n_vals,double* val,long nbits,double reference_value,
        double* scal,double d,double divisor,unsigned char* p,long *bitp) {
  return GRIB_NOT_IMPLEMENTED;
}

#endif

